@@ -6,15 +6,19 @@ import logging
 import os
 import shutil
 import subprocess
-from typing import Any, Iterable, Optional, Tuple, cast
-from urllib.parse import urlparse
+from typing import TYPE_CHECKING, Optional, Tuple, cast
+from urllib.parse import SplitResult, urlparse, urlsplit
 
-from requests import Response
-from requests.auth import AuthBase, HTTPBasicAuth
-from requests.models import PreparedRequest
-from requests.utils import get_netrc_auth
+from httpx import URL, Auth, BasicAuth
 
-from unearth.utils import split_auth_from_netloc, split_auth_from_url
+from unearth.utils import commonprefix, get_netrc_auth, split_auth_from_url
+
+if TYPE_CHECKING:
+    from typing import Any, Callable, Generator, Iterable
+
+    from httpx import Request, Response
+    from requests import Response as RequestsResponse
+    from requests.models import PreparedRequest
 
 KEYRING_DISABLED = False
 
@@ -23,6 +27,14 @@ MaybeAuth = Optional[Tuple[str, Optional[str]]]
 logger = logging.getLogger(__name__)
 
 
+def _expect_argument(func: Callable[..., Any], argname: str) -> bool:
+    """Return True if the function expects the argument."""
+    import inspect
+
+    sig = inspect.signature(func)
+    return argname in sig.parameters
+
+
 class KeyringBaseProvider(metaclass=abc.ABCMeta):
     """Base class for keyring providers."""
 
@@ -43,7 +55,7 @@ class KeyringModuleProvider(KeyringBaseProvider):
     """Keyring provider that uses the keyring module."""
 
     def __init__(self) -> None:
-        import keyring  # type: ignore
+        import keyring
 
         self.keyring = keyring
 
@@ -147,7 +159,9 @@ def get_keyring_auth(url: str | None, username: str | None) -> AuthInfo | None:
         return None
 
 
-class MultiDomainBasicAuth(AuthBase):
+class MultiDomainBasicAuth(Auth):
+    """A multi-domain HTTP basic authentication handler supporting both requests and httpx"""
+
     def __init__(self, prompting: bool = True, index_urls: Iterable[str] = ()) -> None:
         self.prompting = prompting
         self.index_urls = list(index_urls)
@@ -155,22 +169,36 @@ class MultiDomainBasicAuth(AuthBase):
         self._cached_passwords: dict[str, AuthInfo] = {}
         self._credentials_to_save: tuple[str, str, str] | None = None
 
-    def _get_auth_from_index_url(self, netloc: str) -> tuple[MaybeAuth, str | None]:
+    def _get_auth_from_index_url(self, url: str) -> tuple[MaybeAuth, str | None]:
         """Return the extracted auth and the original index URL matching
-        the requested netloc.
+        the requested url. The auth part should be already stripped from the URL.
 
         Returns None if no matching index was found, or if --no-index
         was specified by the user.
         """
-        if not netloc or not self.index_urls:
+        if not url or not self.index_urls:
             return None, None
 
-        for u in self.index_urls:
-            parsed = urlparse(u)
-            auth, index_netloc = split_auth_from_netloc(parsed.netloc)
-            if index_netloc == netloc:
-                return auth, u
-        return None, None
+        target = urlsplit(url.rstrip("/") + "/")
+        candidates: list[SplitResult] = []
+
+        for index in self.index_urls:
+            index = index.rstrip("/") + "/"
+            auth, url_no_auth = split_auth_from_url(index)
+            parsed = urlsplit(url_no_auth)
+            if parsed == target:
+                return auth, index
+
+            if target.netloc == parsed.netloc:
+                candidates.append(urlsplit(index))
+
+        if not candidates:
+            return None, None
+        best_match = max(
+            candidates, key=lambda x: commonprefix(x.path, target.path).rfind("/")
+        )
+        index = best_match.geturl()
+        return split_auth_from_url(index)[0], index
 
     def _get_new_credentials(
         self,
@@ -193,14 +221,22 @@ class MultiDomainBasicAuth(AuthBase):
                 return cast(AuthInfo, auth)
 
         # Find a matching index url for this request
-        index_auth, index_url = self._get_auth_from_index_url(netloc)
+        # XXX: a compatiblity layer to support old function signature
+        if _expect_argument(self._get_auth_from_index_url, "netloc"):
+            index_auth, index_url = self._get_auth_from_index_url(netloc)
+        else:
+            index_auth, index_url = self._get_auth_from_index_url(url)
+
         if index_url:
             logger.debug("Found index url %s", index_url)
 
             # If an index URL was found, try its embedded credentials
-            if index_auth is not None and index_auth[1] is not None:
-                logger.debug("Found credentials in index url for %s", netloc)
-                return cast(AuthInfo, index_auth)
+            if index_auth is not None:
+                if index_auth[1] is not None:
+                    logger.debug("Found credentials in index url for %s", netloc)
+                    return cast(AuthInfo, index_auth)
+                if username is None:
+                    username = index_auth[0]
 
         # Get creds from netrc if we still don't have them
         if allow_netrc:
@@ -259,6 +295,8 @@ class MultiDomainBasicAuth(AuthBase):
 
     def __call__(self, req: PreparedRequest) -> PreparedRequest:
         # Get credentials for this request
+        from requests.auth import HTTPBasicAuth
+
         url, username, password = self._get_url_and_credentials(cast(str, req.url))
         req.url = url
 
@@ -270,14 +308,72 @@ class MultiDomainBasicAuth(AuthBase):
 
         return req
 
+    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
+        url, username, password = self._get_url_and_credentials(str(request.url))
+        request.url = URL(url)
+
+        if username is not None and password is not None:
+            basic_auth = BasicAuth(username, password)
+            request = next(basic_auth.auth_flow(request))
+
+        response = yield request
+
+        if response.status_code != 401:
+            return
+
+        # Query the keyring for credentials:
+        username, password = self._get_new_credentials(
+            url, allow_netrc=False, allow_keyring=True
+        )
+
+        # Prompt the user for a new username and password
+        save = False
+        netloc = response.url.netloc.decode()
+        if password is None:
+            # We are not able to prompt the user so simply return the response
+            if not self.prompting:
+                return
+
+            if _expect_argument(self._prompt_for_password, "username"):
+                username, password, save = self._prompt_for_password(netloc, username)
+            else:
+                username, password, save = self._prompt_for_password(netloc)
+
+        # Store the new username and password to use for future requests
+        self._credentials_to_save = None
+        if username is not None and password is not None:
+            self._cached_passwords[netloc] = (username, password)
+
+            # Prompt to save the password to keyring
+            if save and self._should_save_password_to_keyring():
+                self._credentials_to_save = (netloc, username, password)
+
+        # Add our new username and password to the request
+        basic_auth = BasicAuth(username or "", password or "")
+        request = next(basic_auth.auth_flow(request))
+
+        response = yield request
+        self.warn_on_401(response)
+
+        # On successful request, save the credentials that were used to
+        # keyring. (Note that if the user responded "no" above, this member
+        # is not set and nothing will be saved.)
+        if self._credentials_to_save:
+            self.save_credentials(response)
+
     # Factored out to allow for easy patching in tests
-    def _prompt_for_password(self, netloc: str) -> tuple[str | None, str | None, bool]:
-        username = input(f"User for {netloc}: ")
+    def _prompt_for_password(
+        self, netloc: str, username: str | None = None
+    ) -> tuple[str | None, str | None, bool]:
+        if username is None:
+            username = input(f"User for {netloc}: ")
+            auth = get_keyring_auth(netloc, username)
+            if auth and auth[0] is not None and auth[1] is not None:
+                return (*auth, False)
+        else:
+            logger.info("Username: %s", username)
         if not username:
             return None, None, False
-        auth = get_keyring_auth(netloc, username)
-        if auth and auth[0] is not None and auth[1] is not None:
-            return auth[0], auth[1], False
         password = getpass.getpass("Password: ")
         return username, password, True
 
@@ -287,9 +383,11 @@ class MultiDomainBasicAuth(AuthBase):
             return False
         return input("Save credentials to keyring [y/N]: ") == "y"
 
-    def handle_401(self, resp: Response, **kwargs: Any) -> Response:
+    def handle_401(self, resp: RequestsResponse, **kwargs: Any) -> Response:
         # We only care about 401 response, anything else we want to just
         #   pass through the actual response
+        from requests.auth import HTTPBasicAuth
+
         if resp.status_code != 401:
             return resp
 
@@ -304,12 +402,17 @@ class MultiDomainBasicAuth(AuthBase):
 
         # Prompt the user for a new username and password
         save = False
-        if not username and not password:
+        if password is None:
             # We are not able to prompt the user so simply return the response
             if not self.prompting:
                 return resp
 
-            username, password, save = self._prompt_for_password(parsed.netloc)
+            if _expect_argument(self._prompt_for_password, "username"):
+                username, password, save = self._prompt_for_password(
+                    parsed.netloc, username
+                )
+            else:
+                username, password, save = self._prompt_for_password(parsed.netloc)
 
         # Store the new username and password to use for future requests
         self._credentials_to_save = None
@@ -338,12 +441,12 @@ class MultiDomainBasicAuth(AuthBase):
             req.register_hook("response", self.save_credentials)
 
         # Send our new request
-        new_resp = resp.connection.send(req, **kwargs)  # type: ignore
+        new_resp = resp.connection.send(req, **kwargs)  # type: ignore[attr-defined]
         new_resp.history.append(resp)
 
         return new_resp
 
-    def warn_on_401(self, resp: Response, **kwargs: Any) -> None:
+    def warn_on_401(self, resp: Response | RequestsResponse, **kwargs: Any) -> None:
         """Response callback to warn about incorrect credentials."""
         if resp.status_code == 401:
             logger.warning(
@@ -352,7 +455,9 @@ class MultiDomainBasicAuth(AuthBase):
                 resp.request.url,
             )
 
-    def save_credentials(self, resp: Response, **kwargs: Any) -> None:
+    def save_credentials(
+        self, resp: Response | RequestsResponse, **kwargs: Any
+    ) -> None:
         """Response callback to save credentials on success."""
         keyring = get_keyring_provider()
         assert keyring is not None, "should never reach here without keyring"
